@@ -28,15 +28,17 @@
 #include <unistd.h>
 #include <string.h>
 #include <gnome.h>
+#include <glade/glade.h>
 #include "gok-settings-dialog.h"
 #include "gok-page-keysizespace.h"
 #include "gok-page-actions.h"
 #include "gok-page-feedbacks.h"
 #include "gok-page-accessmethod.h"
 #include "gok-page-wordcomplete.h"
+#include "gok-page-keyboard.h"
 #include "gok-page-accessmethod.h"
+#include "gok-glade-helpers.h"
 #include "gok-data.h"
-#include "support.h"
 #include "gok-log.h"
 #include "main.h"
 #include "gok-gconf-keys.h"
@@ -44,6 +46,9 @@
 /* pointer to the settings dialog window */
 static GtkWidget* m_pWindowSettings;
 
+/* pointer to the settings dialog xml data type */
+static GladeXML* m_pXML;
+
 static gboolean is_locked;
 
 /**
@@ -58,142 +63,27 @@ static gboolean is_locked;
 **/
 gboolean gok_settingsdialog_open (gboolean bShow)
 {
-	GtkWidget* pButtonTry;
-	GtkWidget* pButtonRevert;
-	GtkWidget* pButtonOK;
-	GtkWidget* pButtonCancel;
-	GtkWidget* pButtonHelp;
-	GtkWidget* hbuttonbox;
-	GtkTooltips *tooltips;
-
-	m_pWindowSettings = create_dialogSettings();
-	if (m_pWindowSettings == NULL)
-	{
-		gok_log_x ("Error: Settings dialog not created in gok_settingsdialog_open!\n");
-		return FALSE;
-	}
-
+	GladeXML *xml;
+	
+	xml = gok_glade_xml_new("gok.glade2", "dialogSettings");
+	m_pXML = xml;
+	
+	m_pWindowSettings = glade_xml_get_widget(xml, "dialogSettings");
+	g_assert (m_pWindowSettings != NULL);
+	
 	is_locked = FALSE;  /* don't read gconf here */
 
 	/* initialize all the pages */
-	gok_settings_page_keysizespace_initialize (m_pWindowSettings);
-	gok_page_accessmethod_initialize (m_pWindowSettings);
-	gok_page_actions_initialize (m_pWindowSettings);
-	gok_page_feedbacks_initialize (m_pWindowSettings);
-	gok_page_wordcomplete_initialize (m_pWindowSettings);
+	gok_settings_page_keysizespace_initialize (xml);
+	gok_page_accessmethod_initialize (xml);
+	gok_page_keyboard_initialize (xml);
+	gok_page_actions_initialize (xml);
+	gok_page_feedbacks_initialize (xml);
+	gok_page_wordcomplete_initialize (xml);
 	
 	/* create a backup copy of all the settings */
 	gok_data_backup_settings();
-	
-	tooltips = gtk_tooltips_new ();
-	
-/* add the buttons (OK, CANCEL, APPLY HELP) at the bottom of the dialog */
-/* I'm doing this here because Glade generates only the GTK 1.2 calls and we need GTK 1.3 or greater*/
-
-	/* get the button box that will hold the buttons */
-	hbuttonbox = lookup_widget (m_pWindowSettings, "hboxStockButtons");
-	if (hbuttonbox == NULL)
-	{
-		gok_log_x ("Error: Can't find the button box for settings dialog in gok_settingsdialog_open!\n");
-		return FALSE;
-	}
-
-	gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);
-
-#if GTK_MAJOR_VERSION < 2
-#if GTK_MINOR_VERSION < 3
-/* these calls can be used only on GTK 1.2 and below */
-
-#define EARLY_GTK_VERSION 1
-
-	pButtonTry = gnome_stock_button (GNOME_STOCK_BUTTON_TRY);
-	pButtonRevert = gnome_stock_button (GNOME_STOCK_BUTTON_REVERT);
-	pButtonOK = gnome_stock_button (GNOME_STOCK_BUTTON_OK);
-	pButtonCancel = gnome_stock_button (GNOME_STOCK_BUTTON_CANCEL);
-	pButtonHelp = gnome_stock_button (GNOME_STOCK_BUTTON_HELP);
- 
-#endif /* #if GTK_MINOR_VERSION < 3 */
-#endif /* #if GTK_MAJOR_VERSION < 2 */
-
-#ifndef EARLY_GTK_VERSION
-/* these calls can be used only on GTK 1.3 and above */
-
-	pButtonTry = gtk_button_new_from_stock (GTK_STOCK_JUMP_TO);
-	pButtonRevert = gtk_button_new_from_stock (GTK_STOCK_REVERT_TO_SAVED);
-	pButtonOK = gtk_button_new_from_stock (GTK_STOCK_OK);
-	pButtonCancel = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
-	pButtonHelp = gtk_button_new_from_stock (GTK_STOCK_HELP);
-#endif
-
-	gtk_widget_ref (pButtonHelp);
-	gtk_object_set_data_full (GTK_OBJECT (m_pWindowSettings), "pButtonHelp", pButtonHelp,
-	                            (GtkDestroyNotify) gtk_widget_unref);
-	gtk_widget_show (pButtonHelp);
-	gtk_box_pack_start (GTK_BOX (hbuttonbox), pButtonHelp, FALSE, FALSE, 10);
-#ifndef EARLY_GTK_VERSION
-	gtk_button_box_set_child_secondary (GTK_BUTTON_BOX (hbuttonbox), pButtonHelp, TRUE);
-#endif
-	GTK_WIDGET_SET_FLAGS (pButtonHelp, GTK_CAN_DEFAULT);
-	gtk_tooltips_set_tip (tooltips, pButtonHelp, _("display user help"), NULL);
-
-	gtk_signal_connect (GTK_OBJECT (pButtonHelp), "clicked",
-                      GTK_SIGNAL_FUNC (on_button_help),
-                      NULL);
-	
-	gtk_button_set_label (GTK_BUTTON(pButtonTry), _("_Try"));
-/*	gtk_button_set_use_stock (GTK_BUTTON(pButtonTry), TRUE); */
-
-	gtk_widget_ref (pButtonTry);
-	gtk_object_set_data_full (GTK_OBJECT (m_pWindowSettings), "pButtonTry", pButtonTry,
-	                            (GtkDestroyNotify) gtk_widget_unref);
-	gtk_widget_show (pButtonTry);
-	gtk_box_pack_start (GTK_BOX (hbuttonbox), pButtonTry, FALSE, FALSE, 10);
-	GTK_WIDGET_SET_FLAGS (pButtonTry, GTK_CAN_DEFAULT);
-	gtk_tooltips_set_tip (tooltips, pButtonTry, _("try these settings"), NULL);
-
-	gtk_signal_connect (GTK_OBJECT (pButtonTry), "clicked",
-                      GTK_SIGNAL_FUNC (on_button_try),
-                      NULL);
-
-	gtk_widget_ref (pButtonRevert);
-	gtk_object_set_data_full (GTK_OBJECT (m_pWindowSettings), "pButtonRevert", pButtonRevert,
-	                            (GtkDestroyNotify) gtk_widget_unref);
-	gtk_widget_show (pButtonRevert);
-	gtk_box_pack_start (GTK_BOX (hbuttonbox), pButtonRevert, FALSE, FALSE, 10);
-	GTK_WIDGET_SET_FLAGS (pButtonRevert, GTK_CAN_DEFAULT);
-	gtk_tooltips_set_tip (tooltips, pButtonRevert, _("revert to original settings"), NULL);
-
-	gtk_signal_connect (GTK_OBJECT (pButtonRevert), "clicked",
-                      GTK_SIGNAL_FUNC (on_button_revert),
-                      NULL);
-
-	gtk_widget_ref (pButtonCancel);
-	gtk_object_set_data_full (GTK_OBJECT (m_pWindowSettings), "pButtonCancel", pButtonCancel,
-	                            (GtkDestroyNotify) gtk_widget_unref);
-	gtk_widget_show (pButtonCancel);
-	gtk_box_pack_start (GTK_BOX (hbuttonbox), pButtonCancel, FALSE, FALSE, 10);
-
-	GTK_WIDGET_SET_FLAGS (pButtonCancel, GTK_CAN_DEFAULT);
-	gtk_tooltips_set_tip (tooltips, pButtonCancel, _("cancel"), NULL);
-
-	gtk_signal_connect (GTK_OBJECT (pButtonCancel), "clicked",
-                      GTK_SIGNAL_FUNC (on_button_cancel),
-                      NULL);
-
-	gtk_widget_ref (pButtonOK);
-	gtk_object_set_data_full (GTK_OBJECT (m_pWindowSettings), "pButtonOK", pButtonOK,
-	                            (GtkDestroyNotify) gtk_widget_unref);
-	gtk_widget_show (pButtonOK);
-
-	gtk_box_pack_start (GTK_BOX (hbuttonbox), pButtonOK, FALSE, FALSE, 10);
-	GTK_WIDGET_SET_FLAGS (pButtonOK, GTK_CAN_DEFAULT);
-	gtk_tooltips_set_tip (tooltips, pButtonOK, _("use these settings"), NULL);
-
-	gtk_signal_connect (GTK_OBJECT (pButtonOK), "clicked",
-                      GTK_SIGNAL_FUNC (on_button_ok),
-                      NULL);
-
-	
+			
 	if (bShow == TRUE)
 	{
 		gok_settingsdialog_show ();
@@ -223,15 +113,13 @@ gboolean gok_settingsdialog_show()
 		GOK_GCONF_PREFS_LOCKED, &is_locked);
 	
 	if (was_locked != is_locked) {
-		notebook = lookup_widget (m_pWindowSettings, "notebook1");
-		/* FIXME: magic #'s used - easily fixed later when we libglade-ify */
 		if (!is_locked) {
-			gtk_widget_show(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1));
-			gtk_widget_show(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 3));
+			gtk_widget_show(glade_xml_get_widget(m_pXML, "ActionsTab"));
+			gtk_widget_show(glade_xml_get_widget(m_pXML, "AccessMethodsTab"));
 		}
 		else {
-			gtk_widget_hide(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1));
-			gtk_widget_hide(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 3));
+			gtk_widget_hide(glade_xml_get_widget(m_pXML, "ActionsTab"));
+			gtk_widget_hide(glade_xml_get_widget(m_pXML, "AccessMethodsTab"));
 		}
 	}
 
@@ -271,6 +159,16 @@ GtkWidget* gok_settingsdialog_get_window ()
  	return m_pWindowSettings;
 }
 
+/**
+* gok_settingsdialog_get_xml
+*
+* returns: A pointer to the settings dialog glade xml data type.
+**/
+GladeXML* gok_settingsdialog_get_xml ()
+{
+ 	return m_pXML;
+}
+
 /**
 * gok_settingsdialog_close
 * 
@@ -336,6 +234,10 @@ void on_button_try (GtkButton* button, gpointer user_data)
 	{
 		bDataChanged = TRUE;
 	}
+	if (gok_page_keyboard_apply() == TRUE)
+	{
+		bDataChanged = TRUE;
+	}
 	if (gok_page_feedbacks_apply() == TRUE)
 	{
 		bDataChanged = TRUE;
@@ -352,7 +254,9 @@ void on_button_try (GtkButton* button, gpointer user_data)
 	/* if any settings have changed then refresh the keyboard */
 	if (bDataChanged == TRUE)
 	{
-		gok_main_display_scan_reset();		
+		gok_main_display_scan_reset();
+		/* and post a warning if the new settings use corepointer */
+		gok_main_warn_if_corepointer_mode (_("You appear to be configuring GOK to use \'core pointer\' mode."), FALSE);
 	}
 }
 
@@ -378,6 +282,10 @@ void on_button_revert (GtkButton* button, gpointer user_data)
 	{
 		bDataChanged = TRUE;
 	}
+	if (gok_page_keyboard_revert() == TRUE)
+	{
+		bDataChanged = TRUE;
+	}
 	if (gok_page_accessmethod_revert() == TRUE)
 	{
 		bDataChanged = TRUE;
@@ -429,8 +337,9 @@ void on_button_cancel (GtkButton* pButton, gpointer user_data)
 void on_button_help (GtkButton* pButton, gpointer user_data)
 {
 	GtkWidget* pNotebook;
+	const char* help_link_id = NULL;
 	
-	pNotebook = lookup_widget (m_pWindowSettings, "notebook1");
+	pNotebook = glade_xml_get_widget (m_pXML, "notebook1");
 	if (pNotebook == NULL)
 	{
 		gok_log_x ("Can't find notebook!\n");
@@ -440,29 +349,36 @@ void on_button_help (GtkButton* pButton, gpointer user_data)
 	switch (gtk_notebook_get_current_page (GTK_NOTEBOOK (pNotebook)))
 	{
 		case 0:
-			gnome_help_display ("gok.xml","gok-settings-appearance", NULL);
+			help_link_id = "gok-prefs-appearance";
 			break;
 			
 		case 1:
-			gnome_help_display ("gok.xml", "gok-settings-actions", NULL);
+			help_link_id = "gok-keyboards";
 			break;
 			
 		case 2:
- 			gnome_help_display ("gok.xml", "gok-settings-feedback", NULL);
-			break;
-
-		case 3:
-			gnome_help_display ("gok.xml", "gok-settings-accessmethods",
-                                            NULL);
+			help_link_id = "gok-prefs-Actions";
 			break;
 			
+		case 3:
+			help_link_id = "gok-prefs-Feedback";
+			break;
+
 		case 4:
-			gnome_help_display ("gok.xml", "gok-settings-prediction", NULL);
+			help_link_id = "gok-prefs-Access-Methods";
+			break;
+			
+		case 5:
+			help_link_id = "gok-prefs-Prediction";
 			break;
 			
 		default:
 			break;
 	}
+	if (help_link_id)
+	{
+		gnome_help_display_desktop (NULL, "gok", "gok.xml", help_link_id, NULL);
+	}
 }
 
 /**
@@ -486,6 +402,7 @@ void gok_settingsdialog_backup_settings ()
 {
 	gok_settings_page_keysizespace_backup();
 	gok_page_actions_backup();
+	gok_page_keyboard_backup();
 	gok_page_feedbacks_backup();
 	gok_page_accessmethod_backup();
 	gok_page_wordcomplete_backup();
