@@ -74,10 +74,6 @@ static SPIBoolean gok_spy_mouse_listener (const AccessibleDeviceEvent *event,
 SPIBoolean gok_spy_key_listener (const AccessibleKeystroke *key,
                                  void *user_data);
 
-static GSList*  gok_spy_append_node (GSList *nodes,
-				     Accessible* pAccessible,
-				     AccessibleNodeFlags flags);
-
 static void gok_spy_modifier_listener(const AccessibleEvent *event,
                                       void *user_data);
 
@@ -96,7 +92,9 @@ static gboolean gok_spy_is_visible (Accessible *accessible);
 		
 static void gok_spy_free_nodes (GSList *nodelist);										
 static void gok_spy_resolve_namesakes (GSList *nodelist);
-												
+
+static gboolean gok_spy_is_primary_container (Accessible *accessible);
+									
 /*
  * gok-spy variables
  */
@@ -117,6 +115,8 @@ static Accessible* m_ptheAccessibleWithText = NULL;
 
 static Accessible* theContextMenuAccessible = NULL;
 
+static Accessible* currently_focussed_object = NULL;
+
 static AccessibleEventListener* focusListener;
 static AccessibleEventListener* defunctnessListener;
 static AccessibleDeviceListener* mouseListener;
@@ -134,6 +134,32 @@ static gboolean spyshutdown = FALSE;/* this global could be removed if we have a
 
 #define MAX_CHILD_COUNT 100
 
+
+static void
+gok_spy_set_focussed_object (Accessible *accessible)
+{
+    if (currently_focussed_object) 
+    {
+	gok_spy_accessible_unref (currently_focussed_object);
+    }
+    currently_focussed_object = accessible;
+    if (accessible) 
+    {
+	gok_spy_accessible_ref (accessible);
+    }
+}
+
+/*
+ * gok_spy_get_focussed_object:
+ *
+ * returns: the currently focussed object, or NULL if no object has emitted a focus event.
+ */
+Accessible *
+gok_spy_get_focussed_object (void)
+{
+    return currently_focussed_object;
+}
+
 /* 
  * for debugging
  */
@@ -331,6 +357,10 @@ void gok_spy_open()
 			objectStateListener, "object:state-changed");
 		gok_log ("SPI_registerGlobalEventListener for object:state-changed returned: %d", success);
 
+		success = SPI_registerGlobalEventListener (
+			objectStateListener, "object:children-changed");
+		gok_log ("SPI_registerGlobalEventListener for object:children-changed returned: %d", success);
+
 	}
 	gok_log_leave ();
 }
@@ -414,6 +444,8 @@ void gok_spy_close(void)
 		m_gokSpyOpen = FALSE;
 		
 		gok_spy_stop ();
+
+		gok_spy_free_nodes (_priv_ui_nodes);
 		
 		if (m_ptheWindowAccessible != NULL) gok_spy_accessible_unref(m_ptheWindowAccessible);
 		if (m_ptheAccessibleWithText != NULL) gok_spy_accessible_unref(m_ptheAccessibleWithText);
@@ -436,8 +468,8 @@ void gok_spy_close(void)
 		g_queue_free(eventQueue);
 		eventQueue = NULL;
 	}
-	gok_spy_free_nodes (_priv_ui_nodes);
 
+	gok_log ("Exiting SPI mainloop.\n");
 	SPI_exit ();
 
 	gok_log_leave ();
@@ -557,10 +589,28 @@ gok_spy_node_match (AccessibleNode *node, GokSpySearchType type)
 			return TRUE;
 		break;
 	case GOK_SPY_SEARCH_CHILDREN:
+	case GOK_SPY_SEARCH_COMBO:
 	case GOK_SPY_SEARCH_LISTITEMS:
+	case GOK_SPY_SEARCH_TABLE_CELLS:
 	case GOK_SPY_SEARCH_ALL:
 		return TRUE;
 		break;
+	case GOK_SPY_SEARCH_ACTIONABLE:
+		if (Accessible_isAction (node->paccessible))
+		{
+			return TRUE;
+		}
+		else 
+		{
+		    AccessibleStateSet *state = Accessible_getStateSet (node->paccessible);
+		    if (AccessibleStateSet_contains (state, SPI_STATE_SELECTABLE))
+		    {
+			AccessibleStateSet_unref (state);
+			return TRUE;
+		    }
+		    AccessibleStateSet_unref (state);
+		}
+		break;
 	case GOK_SPY_SEARCH_APPLICATIONS:
 		if (Accessible_isApplication (node->paccessible))
 			return TRUE;
@@ -571,6 +621,49 @@ gok_spy_node_match (AccessibleNode *node, GokSpySearchType type)
 	return FALSE;
 }
 
+/**
+ *
+ * gok_spy_get_editable:
+ *
+ * Returns the first editable-text child of @search_root.
+ * Used in particular to locate the text-entry child of editable comboboxes.
+ **/
+Accessible *
+gok_spy_get_editable (Accessible *search_root)
+{
+    Accessible *child = NULL;
+
+    /* TODO: complete the logic below, to search the whole tree from the root, 
+       not just the immediate children */
+    gok_spy_accessible_ref (search_root);
+    while (search_root) 
+    {
+	int i = 0;
+	int child_count = Accessible_getChildCount (search_root);
+	while (i < child_count) 
+	{
+	    AccessibleRole role = SPI_ROLE_INVALID;
+	    child = Accessible_getChildAtIndex (search_root, i);
+	    if (child) 
+	    {
+		gok_spy_accessible_implicit_ref (child);
+		if (Accessible_isEditableText (child)) 
+		{
+		    return child;
+		}
+		else 
+		{
+		    gok_spy_accessible_unref (child);
+		}
+	    }
+	    ++i;
+	}
+	gok_spy_accessible_unref (search_root);
+	search_root = NULL;
+    }
+    return NULL;
+}
+
 /**
  *
  * gok_spy_get_list_parent:
@@ -583,7 +676,6 @@ Accessible *
 gok_spy_get_list_parent (Accessible *search_root)
 {
     Accessible *child = NULL;
-
     gok_spy_accessible_ref (search_root);
     while (search_root) 
     {
@@ -596,7 +688,9 @@ gok_spy_get_list_parent (Accessible *search_root)
 		gok_spy_accessible_implicit_ref (child);
 		role = Accessible_getRole (child);
 		if (role == SPI_ROLE_LIST_ITEM
-		    || role == SPI_ROLE_MENU_ITEM)
+		    || role == SPI_ROLE_MENU_ITEM
+		    || role == SPI_ROLE_RADIO_MENU_ITEM
+		    || role == SPI_ROLE_CHECK_MENU_ITEM)
 		{
 		    gok_spy_accessible_unref (child);
 		    return search_root;
@@ -617,6 +711,139 @@ gok_spy_get_list_parent (Accessible *search_root)
     return NULL;
 }
 
+static gchar*
+gok_spy_concatenate_child_names (Accessible *parent)
+{
+	gint num_children = 0;
+	gchar *tmp = NULL, *name = NULL;
+	if (parent) 
+	{
+		gint i;
+		num_children = Accessible_getChildCount (parent);
+		for (i = 0; i < num_children; ++i) 
+		{
+			Accessible *child = Accessible_getChildAtIndex (parent, i);
+			if (child ) 
+			{
+				gchar *childname = Accessible_getName (child);
+				if (childname && (strlen (childname) > 0)) 
+				{
+					if (tmp)
+					{
+						name = g_strconcat (tmp, "", childname, NULL);
+						g_free (tmp);
+					}
+					else
+					{
+						name = g_strdup (childname);
+					}
+					SPI_freeString (childname);
+					tmp = name;
+				}
+			}
+		}
+	}
+	return name;
+}
+
+/**
+ * 
+ * gok_spy_get_table_nodes:
+ * Returns a list of nodes which are children of a table, such that
+ * there is exactly one child per currently visible cell 
+ * plus one child per actionable table header.
+ *
+ **/
+GSList *
+gok_spy_get_table_nodes (Accessible *search_root)
+{
+	GSList *nodes = NULL;
+	AccessibleNodeFlags flags;
+	AccessibleTable *table;
+	Accessible *child, *header, *first_child, *last_child;
+	AccessibleComponent *component;
+	long row_count, col_count;
+	long i, first_row, last_row, first_index = 0, last_index = -1;
+	long x, y, width, height;
+
+	g_assert (Accessible_isTable (search_root));
+	table = Accessible_getTable (search_root);
+	flags.value = 0xff;
+	col_count = AccessibleTable_getNColumns (table);
+	if (col_count > 1)
+	{
+		for (i = 0; i < col_count; ++i)
+		{
+			header = AccessibleTable_getColumnHeader (table, i);
+			if (header)
+			{
+				gok_spy_accessible_implicit_ref (header);
+				nodes = gok_spy_append_node (nodes, header, flags);
+				gok_spy_accessible_unref (header);
+			}
+		}
+	}
+
+	component = Accessible_getComponent (table);
+	if (!component) return NULL;
+
+	AccessibleComponent_getExtents (component, &x, &y, &width, &height, SPI_COORD_TYPE_SCREEN);
+	first_child = AccessibleComponent_getAccessibleAtPoint (component, x, y, SPI_COORD_TYPE_SCREEN);
+	last_child = AccessibleComponent_getAccessibleAtPoint (component, x + width, y + height, SPI_COORD_TYPE_SCREEN);
+	AccessibleComponent_unref (component);
+
+	if (first_child) 
+		first_index = Accessible_getIndexInParent (first_child);
+	if (last_child) 
+		last_index = Accessible_getIndexInParent (last_child);
+	Accessible_unref (first_child);
+	Accessible_unref (last_child);
+
+	gok_log ("first row index %d; last row index %d", first_index, last_index);
+
+	row_count = AccessibleTable_getNRows (table);
+	first_row = AccessibleTable_getRowAtIndex (table, first_index);
+	if (last_index >= 0) 
+		last_row = AccessibleTable_getRowAtIndex (table, last_index);
+	else 
+		last_row = row_count - 1;
+
+	last_row = MIN (last_row, row_count - 1);
+	first_row = MAX (first_row, 0);
+	
+	gok_log ("adding nodes for rows %d through %d (inclusive)\n", first_row, last_row);
+
+	for (i = first_row; i <= last_row; ++i)
+	{
+		gint n = 0;
+		/* make sure the row cell we choose is not anonymous */
+		/* assignment operator on child ("=" vs "==") is intentional */
+		while ((child = AccessibleTable_getAccessibleAt (table, i, n)) && (n < col_count))
+		{
+			char *name = Accessible_getName (child);
+			char *desc = Accessible_getDescription (child);
+			gboolean has_name = (name && (strlen (name) > 0)), has_desc = (desc && (strlen (desc) > 0));
+
+			gok_spy_accessible_implicit_ref (child);
+			if (name) SPI_freeString (name);
+			if (desc) SPI_freeString (desc);
+			/* we make exception for compound cells, which will scavenge a name from their children */
+			if (has_name || has_desc || (Accessible_getChildCount (child) > 0))
+			{
+				nodes = gok_spy_append_node (nodes, child, flags);
+				gok_spy_accessible_unref (child);
+			}
+			else 
+			{
+ 			        gok_log ("ignoring node at row %d, col %d", i, n);
+				gok_spy_accessible_unref (child);
+			}
+			++n;
+		}
+	}
+	return nodes;
+}
+
 /**
  * 
  * gok_spy_get_children:
@@ -648,6 +875,57 @@ gok_spy_get_children (Accessible *search_root)
 	return nodes;
 }
 
+/**
+ * gok_spy_get_actionable_descendants:
+ * Returns a list of actionable or selectable descendants of #accessible,
+ * including #accessible itself.
+ **/
+GSList *
+gok_spy_get_actionable_descendants (Accessible *accessible, GSList *nodes)
+{
+    gint child_count, i, max_children = 20;
+    Accessible *child;
+    AccessibleNodeFlags flags;
+    GSList *ret_nodes;
+    gboolean is_selection = Accessible_isSelection (accessible);
+    flags.value = 0;
+    flags.data.is_ui = TRUE;
+
+    if (Accessible_isAction (accessible))
+    {
+	ret_nodes = gok_spy_append_node (nodes, accessible, flags);
+    }
+    else
+    {
+	ret_nodes = nodes;
+    }
+    child_count = Accessible_getChildCount (accessible);
+    for (i = 0; i < child_count && i < max_children; ++i) 
+    {
+	AccessibleStateSet *stateset;
+	child = Accessible_getChildAtIndex (accessible, i);
+	stateset = Accessible_getStateSet (child);
+	if (Accessible_isAction (child) && AccessibleStateSet_contains (stateset, SPI_STATE_SHOWING))
+	{
+	    ret_nodes = gok_spy_append_node (ret_nodes, child, flags);
+	}
+	else if (is_selection)
+	{
+	    if (AccessibleStateSet_contains (stateset, SPI_STATE_SELECTABLE))
+	    {
+		ret_nodes = gok_spy_append_node (ret_nodes, child, flags);
+	    }
+	}
+	else
+	{
+	    ret_nodes = gok_spy_get_actionable_descendants (child, ret_nodes);
+	}
+	AccessibleStateSet_unref (stateset);
+	Accessible_unref (child);
+    }
+    return ret_nodes;
+}
+
 /** 
 * gok_spy_get_list 
 * 
@@ -694,6 +972,7 @@ AccessibleNode* gok_spy_refresh( AccessibleNode* plist)
 void gok_spy_free_nodes (GSList *nodes)
 {
 	GSList *nodelist = nodes;
+	gok_log_enter();
 	while (nodelist) 
 	{
 		AccessibleNode *node = nodelist->data;
@@ -707,6 +986,7 @@ void gok_spy_free_nodes (GSList *nodes)
 		nodelist = nodelist->next;
 	}
 	g_slist_free (nodes);
+	gok_log_leave();
 }
 
 /** 
@@ -728,6 +1008,7 @@ gboolean gok_spy_check_queues(void)
 {
 	EventNode* en = NULL;
 	char *s;
+	GokKeyboard *current_kbd;
 
 	gok_log_enter();
 	
@@ -761,7 +1042,7 @@ gboolean gok_spy_check_queues(void)
 		{
 		case GOKSPY_FOCUS_EVENT:
 			/* TODO: remove obsolete focus events from queue since process
-			   accessible is expensive. */
+			   focus is expensive. */
 			gok_log("GOKSPY_FOCUS_EVENT");
 			gok_spy_process_focus (en->event->source);
 			break;
@@ -790,6 +1071,13 @@ gboolean gok_spy_check_queues(void)
 			m_ptheAccessibleWithText = NULL;
 			((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			break;
+		case GOKSPY_CONTAINER_EVENT:
+			gok_log("GOKSPY_CONTAINER_EVENT");
+			current_kbd = gok_main_get_current_keyboard ();
+			gok_spy_update_component_list (gok_main_get_foreground_window_accessible (), 
+			  current_kbd->flags);
+			gok_main_ds (current_kbd);
+			break;
 		case GOKSPY_WINDOW_DEACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_DEACTIVATE_EVENT");
 
@@ -938,7 +1226,24 @@ gok_spy_is_menu_role (AccessibleRole role)
 
 
 static gboolean
-gok_spy_is_ui (Accessible *accessible, AccessibleRole role)
+gok_spy_is_selectable_child (Accessible *accessible, Accessible *parent)
+{
+    if (parent && Accessible_isSelection (parent))
+    {
+/* see bug #153638
+	gboolean retval;
+	AccessibleStateSet *stateset = Accessible_getStateSet (accessible);
+	retval = AccessibleStateSet_contains (stateset, SPI_STATE_SELECTABLE);
+	AccessibleStateSet_unref (stateset);
+	return retval;
+*/
+	return TRUE;
+    }
+    return FALSE;
+}
+
+static gboolean
+gok_spy_is_ui (Accessible *accessible, Accessible *parent, AccessibleRole role)
 {
 	gboolean interesting = FALSE;
 
@@ -951,13 +1256,18 @@ gok_spy_is_ui (Accessible *accessible, AccessibleRole role)
 	case SPI_ROLE_PAGE_TAB:
 	case SPI_ROLE_TOGGLE_BUTTON:
 	case SPI_ROLE_SLIDER:
+	case SPI_ROLE_SCROLL_BAR:
+        case SPI_ROLE_LIST:
 		interesting = TRUE;
 		break;
 	default:
 		if (Accessible_isEditableText (accessible) ||
 		    Accessible_isHypertext (accessible) ||
-		    Accessible_isAction (accessible))
-			interesting = TRUE;
+		    Accessible_isTable (accessible) ||
+		    (!gok_spy_is_menu_role (role) && 
+		     (Accessible_isAction (accessible) ||
+		      gok_spy_is_selectable_child (accessible, parent))))
+		    interesting = TRUE;
 		gok_log ("checking for interesting interfaces... %s\n", interesting ? "yes" : "no");
 		gok_log ("is icon: %s\n", role == SPI_ROLE_ICON ? "yes" : "no");
 		break;
@@ -1000,7 +1310,6 @@ gok_spy_distinguish_node_name (AccessibleNode *node)
 	if (node->paccessible) 
 	    relations = Accessible_getRelationSet(node->paccessible);
 	if (relations != NULL) {
-#ifdef NEVER		
 	 while (relations[i]) {
 		if (!distinguished) {
 			type = AccessibleRelation_getRelationType(relations[i]);
@@ -1011,10 +1320,10 @@ gok_spy_distinguish_node_name (AccessibleNode *node)
 						AccessibleRelation_getTarget(relations[i], j);
 					if (labelAccessible != NULL) {
 						s = Accessible_getName (labelAccessible);
-						if (s != NULL) {
+						/* don't prepend if we've already used this label as GokButton name */
+						if (s != NULL && strcmp (s, node->pname)) {
 							if ( strlen (s) != 0 )  {
-								gchar buffer[64];
-								buffer[63] = '\0';
+								gchar buffer[64] = "";
 								if (node->pname) 
 								{
 									if (strlen (node->pname)) 
@@ -1023,6 +1332,7 @@ gok_spy_distinguish_node_name (AccessibleNode *node)
 									}
 									g_free (node->pname); 
 								}
+								buffer[63] = '\0'; /* ensure null termination */
 								node->pname = g_strconcat (s, " ", buffer, NULL);
 								SPI_freeString (s);
 								Accessible_unref (labelAccessible);
@@ -1040,7 +1350,6 @@ gok_spy_distinguish_node_name (AccessibleNode *node)
 		AccessibleRelation_unref(relations[i]);
 		i++;
 	 }
-#endif
 	 g_free (relations);
 	}
 }
@@ -1152,11 +1461,15 @@ GSList* gok_spy_append_node (GSList* nodes,
 	AccessibleNode* pNewNode = NULL;
 	AccessibleRelation** relations = NULL;
 	Accessible* targetAccessible = NULL;
+	AccessibleRole role;
 	AccessibleRelationType type = -1;
+	AccessibleStateSet *states;
+
 	char* pName = NULL, *s;
 	int i = 0;
 	int j = 0;
 	int ntargets = 0;
+	int maxloops = 100;  /* used to gaurd against infinite loops */
 
 	gok_log_enter();
 
@@ -1240,19 +1553,22 @@ GSList* gok_spy_append_node (GSList* nodes,
 			Accessible *parent = Accessible_getParent (pAccessible);
 			/* if there are no siblings, we can use an ancestor's name */
 			gok_log("checking ancestor names.");
+			maxloops = 7; /* ignore great great great great great great grandparent */
 			while (parent && Accessible_getChildCount (parent) == 1) 
 			{
 				char *tmp;
 				Accessible *tmp_parent = parent;
+				if (maxloops-- < 1) { break; }  /* guard against erroneous cyclic parent/child relations */
 				tmp = Accessible_getName (tmp_parent); 
+				gok_log ("getting parent name: parent %x, name %s, %d more loops allowed.\n", tmp_parent, tmp ? tmp : "", maxloops);
 				if (tmp != NULL) 
 				{
 				    if (strlen (tmp) > 0) 
 				    {
 					pName = g_strdup (tmp);
+					SPI_freeString (tmp);
 					break;
 				    }
-				    SPI_freeString (tmp);
 				}
 				parent = Accessible_getParent (tmp_parent);
 				Accessible_unref (tmp_parent);
@@ -1283,6 +1599,114 @@ GSList* gok_spy_append_node (GSList* nodes,
 				SPI_freeString (s);
 			}
 		}
+
+		if (pName == NULL) 
+		{
+		    role = Accessible_getRole (pAccessible);
+		}
+
+		/* 
+		 * Role "TABLE-CELL" is used for 'generic' table children. 
+		 * These children's behavior is governed more by the fact that they are
+		 * children of a table than by their class-specific characteristics (unlike,
+		 * for instance, ROLE_CHECKBUTTON or ROLE_PUSH_BUTTON objects which may be
+		 * stored in tables).  Some table cells are composite objects, and we should
+		 * attempt to create a name string from the concatenation of their children 
+		 * when possible.  In the case of GtkTreeviews in particular, where row-selection
+		 * is the dominant user activity, this is especially appropriate.
+		 */
+		if (pName == NULL && role == SPI_ROLE_TABLE_CELL) {
+			pName = gok_spy_concatenate_child_names (pAccessible);
+		}
+		else if (pName == NULL && role == SPI_ROLE_SCROLL_BAR) {
+		    states = Accessible_getStateSet (pAccessible);
+		    if (AccessibleStateSet_contains (states, SPI_STATE_VERTICAL))
+		    {
+			/* translators: abbreviated version of "Vertical Scrollbar" */
+			pName = g_strdup (_("V Scrollbar"));
+		    }
+		    else
+		    {
+			/* translators: abbreviated version of "Horizontal Scrollbar" */
+			pName = g_strdup (_("H Scrollbar"));
+		    }
+		    AccessibleStateSet_unref (states);
+		}
+		else if (pName == NULL) { /* desperate attempt to reuse singleton-child's name */
+			Accessible *tmp_parent = pAccessible;
+			gok_log("no description... checking for singleton child with name.");
+			maxloops = 3;
+			Accessible_ref (tmp_parent);
+			while (tmp_parent && Accessible_getChildCount (tmp_parent) == 1 && maxloops) 
+			{
+				Accessible *child;
+				child = Accessible_getChildAtIndex (tmp_parent, 0);
+				if (child) 
+				{
+					s = Accessible_getName (child);
+					if (s)
+					{
+						gint len = strlen (s);
+						if (len > 0) {
+							pName = g_strndup (s, len);
+							Accessible_unref (tmp_parent);
+							tmp_parent = NULL;
+						}
+						SPI_freeString (s);
+					}
+				}
+				if (tmp_parent) 
+				{
+					Accessible_unref (tmp_parent);
+					tmp_parent = child;
+				}
+				--maxloops;
+			}
+		}
+		if (pName == NULL && Accessible_isTable (pAccessible)) {
+		        /* Tables are so important we must include them somehow */
+			AccessibleTable *table = Accessible_getTable (pAccessible);
+			Accessible *caption = AccessibleTable_getCaption (table);
+			if (caption) 
+			{
+				char *caption_string = Accessible_getName (caption);
+				if (caption_string) 
+				{
+					pName = g_strndup (caption_string, 20);
+					SPI_freeString (caption_string);
+				}
+			}
+			if (pName == NULL && AccessibleTable_getNColumns (table) == 1) 
+			{
+				char *desc = AccessibleTable_getColumnDescription (table, 0);
+				if (desc) 
+				{
+					pName = g_strndup (desc, 20);
+					SPI_freeString (desc);
+				}
+				else
+				{
+					Accessible *header;
+					header = AccessibleTable_getColumnHeader (table, 0);
+					if (header) 
+					{
+						desc = Accessible_getName (header);
+						if (desc) 
+						{
+							pName = g_strndup (desc, 20);
+							SPI_freeString (desc);
+						}
+						Accessible_unref (header);
+					}
+				}
+			}
+			if (pName == NULL)
+			{
+			        /* translators: "table" as in row/column data structure */
+				pName = g_strdup (_("Table"));
+			}
+		}			
+
 		if (pName != NULL) {
 			gok_log("we have a name: [%s]",pName);
 			/* create a new node */
@@ -1356,7 +1780,6 @@ GSList* gok_spy_append_node (GSList* nodes,
 				nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
 			}
 		}
-			
 	}
 	
 	gok_log_leave();
@@ -1374,6 +1797,7 @@ void gok_spy_focus_listener (const AccessibleEvent *event, void *user_data)
 	EventNode* en; 
 	gok_log_enter();
 	AccessibleEvent_ref (event);
+	gok_spy_set_focussed_object (event->source);
 	en = (EventNode*)g_malloc(sizeof(EventNode));
 	en->event = event;
 	en->type = GOKSPY_FOCUS_EVENT;
@@ -1421,6 +1845,15 @@ gok_spy_keymap_listener (void)
 	return;
 }
 
+
+gboolean
+gok_spy_is_primary_container (Accessible *accessible)
+{
+    AccessibleRole role;
+    if (accessible) role = Accessible_getRole (accessible);
+    return (role == SPI_ROLE_HTML_CONTAINER || role == SPI_ROLE_SCROLL_PANE);
+}
+
 /** 
 * gok_spy_defunctness_listener 
 *
@@ -1439,6 +1872,15 @@ void gok_spy_defunctness_listener (const AccessibleEvent *event, void *user_data
 		gok_spy_add_idle_handler ();
 		gok_log ("Our target window is now defunct!");
 	}
+	else if (gok_spy_is_primary_container (event->source)) 
+	{
+		en = (EventNode*) g_malloc(sizeof(EventNode));
+		en->event = NULL;
+		en->type = GOKSPY_CONTAINER_EVENT;
+		g_queue_push_tail ( eventQueue, en );
+		gok_spy_add_idle_handler ();
+		gok_log ("A primary UI container is now defunct!");
+	}
 	
 	gok_log_leave();
 }
@@ -1477,25 +1919,33 @@ gok_spy_window_deactivate_listener(const AccessibleEvent *event,
 	gok_log_leave ();
 }
 
-
 static void
 gok_spy_object_state_listener(const AccessibleEvent *event,
                                  void *user_data)
 {
 	EventNode *en;
+	gboolean is_container_add = !strcmp (event->type, "object:children-changed:add");
+
 	gok_log_enter ();
 
+	if (is_container_add)
+	{
+	    if (!event->source || !gok_spy_is_primary_container (event->source))
+	    {
+		gok_log_leave ();
+		return;
+	    }
+	}
 	AccessibleEvent_ref (event);
 	en = (EventNode*)g_malloc (sizeof (EventNode));
 	en->event = event;
-	en->type = GOKSPY_STATE_EVENT;
+	en->type = is_container_add ? GOKSPY_CONTAINER_EVENT : GOKSPY_STATE_EVENT;
 	g_queue_push_tail (eventQueue, en);
 	gok_spy_add_idle_handler ();
 
 	gok_log_leave ();
 }
 
-
 void
 gok_spy_set_context_menu_accessible (Accessible *accessible, int action_index)
 {
@@ -1794,16 +2244,19 @@ gok_spy_get_toolbar_item (Accessible *toolbar_child)
 	 * ROLE_TOOLBAR containers can have ROLE_PANEL or similarly useless children;
 	 * we need to drill down to the useful thingy i.e. pushbutton, etc.
 	 */
-	Accessible *parent = toolbar_child;
-	Accessible_ref (parent);
-	while (parent) 
+	Accessible *child = toolbar_child, *tmp;
+	Accessible_ref (child);
+	while (child) 
 	{
-		Accessible *child = Accessible_getChildAtIndex (parent, 0);
-		Accessible_unref (parent);
-		if (child && Accessible_isAction (child)) 
+		if (Accessible_isAction (child)) 
+		{
 			return child;
-		else 
-			parent = child;
+		}
+		else {
+			tmp = child;
+			child = Accessible_getChildAtIndex (child, 0);
+			Accessible_unref (tmp);
+		}
 	}
 	return NULL;
 }
@@ -1823,7 +2276,7 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 
 	ui_flags.value = 0;
 	parent = rootAccessible;
-	
+
 	if (gok_spy_worth_searching (parent) == FALSE) 
 	{
 		gok_log ("not worth searching.");
@@ -1833,18 +2286,26 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 	parent_role = Accessible_getRole (parent);
 	child_count = Accessible_getChildCount (parent);
 	/* MAX_CHILD_COUNT is a bit of a kludge and will cause certain table items to be unreachable via UI Grab */
-	for (i = 0; i < child_count && i < MAX_CHILD_COUNT; ++i)
+	if (child_count > MAX_CHILD_COUNT)
+		child_count = MAX_CHILD_COUNT;
+	for (i = 0; i < child_count; ++i)
 	{
 		AccessibleNodeFlags flags;
 		flags.value = 0;
 		child = Accessible_getChildAtIndex (parent, i);
-		gok_spy_accessible_implicit_ref (child);
 		if (child)
 		{
+			gok_spy_accessible_implicit_ref (child);
 			flags.value = 0;
 			role = Accessible_getRole (child);
 			if (keyboard_ui_flags.data.menus && gok_spy_is_menu_role (role))
 			{
+				if (parent_role == SPI_ROLE_MENU_BAR &&
+				    !gok_spy_is_visible (child))
+				{
+					gok_spy_accessible_unref (child);
+					continue;
+				}
 				ui_flags.data.menus = flags.data.is_menu = TRUE;
 			}
 			else if (keyboard_ui_flags.data.toolbars && 
@@ -1854,11 +2315,23 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 				if (keyboard_ui_flags.data.gui) 
 					flags.data.is_ui = TRUE; /* toolbar items are ui elements too */
 				ui_flags.data.toolbars = flags.data.is_toolbar_item = TRUE;
-				_priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, toolbar_item, flags);
+				if (toolbar_item) 
+				{
+					_priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, toolbar_item, flags);
+					Accessible_unref (toolbar_item);
+				}
+				else
+				{
+				    ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags).value;
+				}
 			}
-			else if (keyboard_ui_flags.data.gui && !flags.data.is_menu && gok_spy_is_ui (child, role))
+			if ((keyboard_ui_flags.data.gui || keyboard_ui_flags.data.editable_text) && 
+			    !flags.data.is_menu && !flags.data.is_toolbar_item && gok_spy_is_ui (child, parent, role))
 			{
-				ui_flags.data.gui = flags.data.is_ui = TRUE;
+				if (keyboard_ui_flags.data.gui) 
+				{
+				    ui_flags.data.gui = flags.data.is_ui = TRUE;
+				}
 				if (keyboard_ui_flags.data.editable_text && Accessible_isEditableText (child)) 
 				{
 					ui_flags.data.editable_text = TRUE;
@@ -1884,8 +2357,6 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 		}
 	}	
 
-	gok_spy_resolve_namesakes (_priv_ui_nodes);
-
 	/* one last check for context menus on the currently focussed object */
 	if (keyboard_ui_flags.data.menus)
 	{
@@ -1919,6 +2390,8 @@ gok_spy_update_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 	gok_spy_free_nodes (_priv_ui_nodes);
 	_priv_ui_nodes = NULL;
 	ui_flags = gok_spy_search_component_list (rootAccessible, keyboard_ui_flags);
+	/* _priv_ui_nodes is now re-assigned, by above call. */
+	gok_spy_resolve_namesakes (_priv_ui_nodes);
 
 	gok_log_leave ();
 
